@@ -21,7 +21,7 @@ OU1d.loglik <- function(trait, phy, model=c("OUrandomRoot","OUfixedRoot"), param
   
   ## User defined parameters
   if (is.null(parameters)) { parameters = parameters.default } else { 
-    if (class(parameters)!= "list") {
+    if (!inherits(parameters, "list")) {
       stop("please specify parameters as a list().")
     } else {
       specified <- !c(is.null(parameters$ancestral.state), 
